// -*- C++ -*-
//
//  This file is part of CBNSIM, the Content-Based Networking
//  Simulation Library.
//
//  Authors: See the file AUTHORS for full details.
//
//  Copyright (C) 2002-2004 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: Node.cc,v 1.19 2005/08/11 15:11:34 hallcp Exp $
//
#include "cbnsim/Node.h"
#include "cbnsim/NodeFailPacket.h"
#include "cbnsim/NodeRecoverPacket.h"
#include "cbnsim/SendMessagePacket.h"
#include "cbnsim/SetPredicatePacket.h"
#include "cbnsim/logging.h"

using namespace cbnsim;
using namespace ssim;
using namespace std;

ProcessId Node :: activate()
{
  CBNASSERT( m_if_table.empty(),
	     "node " << id() << " interface table not empty upon activation" );
  CBNASSERT( m_delay_table.empty(),
	     "node " << id() << " delay table not empty upon activation" );
  CBNASSERT( m_pid_map.empty(),
	     "node " << id() << " pid map not empty upon activation" );

  ProcessId p = ProcessWithPId::activate();

  m_if_table.push_back( p );
  m_pid_map.insert( make_pair( p, 0 ) );
  m_delay_table.push_back( 0 );

  return p;
}

void Node :: add_neighbor( ProcessId nid, Time length )
{
  CBNDEBUG( "add_neighbor(id=" << id() << ",pid=" << pid() << ") = " << nid << " with length " << length );

  m_if_table.push_back( nid );
  m_pid_map.insert( make_pair( nid, (m_if_table.size() - 1) ) );
  m_delay_table.push_back( length );
}

bool Node::broadcast_packet( const Packet *packet, Time lag, bool count_all )
{
    unsigned int size = m_delay_table.size();
    // don't want to send to zero since that is the localhost
    for( unsigned int i = 1; i < size; ++i ) {
      issue_packet( i, packet, lag, count_all || i == 1 );
    }

    m_power_left -= SEND_POWER_DRAIN;
    if(m_power_left < 0) {
	node_fail();
	return false;
    }

    return true;
}

void Node :: issue_packet( unsigned int ifid, const Packet* packet, Time lag, bool count )
{
  CBNASSERT( ifid < m_delay_table.size(),
	     "interface id (" << ifid << ") out of range for node: " << id() );
  if( count )
    {
      if( packet->is_control() )
	{
	  m_ctrl_packets++;
	}
      else
	{
	  m_data_packets++;
	}
    }
  Sim::signal_event( m_if_table[ifid], packet, (m_delay_table[ifid] + lag) );
}

void Node :: issue_packet( ProcessId pid, const Packet* packet, Time lag )
{
  CBNASSERT( m_pid_map.find( pid ) != m_pid_map.end(),
	     "node " << id() << " attempting to issue packet to non-neighboring process: " << pid );
  issue_packet( m_pid_map[pid], packet, lag );
}

void Node :: process_event( const Event* evt )
{
  const NodeFailPacket* nfp;
  const NodeRecoverPacket* nrp;
  const SendMessagePacket* smp;
  const SetPredicatePacket* spp;

  // handle the "special" events coming from the driver
  if( (nfp = dynamic_cast<const NodeFailPacket*>( evt )) )
    {
      node_fail();
    }
  else if( (nrp = dynamic_cast<const NodeRecoverPacket*>( evt )) )
  {
      node_recover();
  }
  else if( (smp = dynamic_cast<const SendMessagePacket*>( evt )) )
    {
// - but no "packet" is recved here, this is a enternal notification
//	m_power_left -= RECV_POWER_DRAIN;
	if(m_power_left >= 0)
	    send_message( smp->mid(), smp->message() );
    }
  else if( (spp = dynamic_cast<const SetPredicatePacket*>( evt )) )
    {
// - but no "packet" is recved here, this is a enternal notification
//	m_power_left -= RECV_POWER_DRAIN;
	if(m_power_left >= 0)
	    set_predicate( spp->predicate(), spp->msg_rate(), spp->is_predicate_null());
    }
  else
    {
      const Packet* pkt = dynamic_cast<const Packet*>( evt );
      
      CBNASSERT( pkt, "Node::process_event received non-Packet event" );
      CBNASSERT( m_pid_map.find( pkt->sender() ) != m_pid_map.end(),
		 "node " << id()
		 << " received packet from non-neigboring process: "
		 << pkt->sender() );
      process_packet( m_pid_map[pkt->sender()], *pkt );
    }

  if(m_power_left < 0) {
      node_fail();
  }
}

unsigned int Node :: p2i( ProcessId pid ) const
{
  map<ProcessId,unsigned int>::const_iterator itr = m_pid_map.find( pid );
  CBNASSERT( itr != m_pid_map.end(),
	     "node " << id() << " requesting interface for non-neighboring process: " << pid );
  return (*itr).second;
}

